-- Sonic Heroes DFF Import v1.0b
-- Adapted from GTA3 DFF importer created by Lance
-- Modified by Nemesis and Milamber
-- 10/10/2003

utility DFFImport "Sonic Heroes DFF Import v1.0b" width:162 height:98
(
	local rwNONE          = 0
	local rwDATA          = 1
	local rwSTRING        = 2
	local rwEXTENSION     = 3
	local rwTEXTURE       = 6
	local rwMATERIAL      = 7
	local rwMATERIALLIST  = 8
	local rwFRAMELIST     = 14
	local rwGEOMETRY      = 15
	local rwCLUMP         = 16
	local rwATOMIC        = 20
	local rwGEOMETRYLIST  = 26
	local rwUNKNOWN       = 288
	local rwMATERIALSPLIT = 1294
	local rwFRAME         = 39056126
	local rwGEO_EXT_UNK   = 261
	
	local rwOBJECT_VERTEX_UV     = 4
	local rwOBJECT_VERTEX_COLOR  = 8
	local rwOBJECT_VERTEX_NORMAL = 16

	struct s_Geometry ( vertData, faceData, normData, texcData, colorData )
	
	local frameName
	local frameMatrix
	local frameParent
	local geomData
	local frameMesh
	
	local ProcessChunk


	group "Options"
	(
		spinner spnScale "Scale:" range:[0.1,1000,10] type:#float scale:0.1
		checkbox chkDebug "Show Debug Info" checked:false
	)
	button btnImportNow "Import..." width:100
	
	fn CreateAtomic geoID frameID =
	(
		local myMesh
		if (geomData[geoID].texcData.count > 0) then
		(
			myMesh = Mesh \
				name:     (frameName[frameID] as string) \
				vertices: geomData[geoID].vertData \
				faces:    geomData[geoID].faceData \
				tverts:   geomData[geoID].texcData
		)
		else
		(
			myMesh = Mesh \
				name:     (frameName[frameID] as string)\
				vertices: geomData[geoID].vertData \
				faces:    geomData[geoID].faceData
		)

		if (geomData[geoID].normData.count > 0) then
			for i = 1 to geomData[geoID].normData.count do
				setNormal myMesh i geomData[geoID].normData[i]

		if (geomData[geoID].texcData.count > 0) then
		(
			buildTVFaces myMesh
			for i = 1 to geomData[geoID].faceData.count do
				setTVFace myMesh i geomData[geoID].faceData[i]
		)

		frameMesh[frameID] = myMesh
	)


	fn CreateDummies =
	(
		for i = 1 to frameName.count do
			if (frameMesh[i] == undefined) then
				frameMesh[i] = Dummy name:frameName[i] boxsize:([.05,.05,.05] * spnScale.value)
	)


	fn ReadVec2 fin =
	(
		local x = ReadFloat fin
		local y = ReadFloat fin
		return (point3 x y 0)
	)


	fn ReadVec3 fin =
	(
		local x = ReadFloat fin
		local y = ReadFloat fin
		local z = ReadFloat fin
		if x == undefined then 
			return (point3 0.0 0.0 0.0)
		if y == undefined then 
			return (point3 0.0 0.0 0.0)
		if z == undefined then 
			return (point3 0.0 0.0 0.0)
		return (point3 x y z)
	)


	fn ReadMatrix fin =
	(
		local r1 = ReadVec3 fin
		local r2 = ReadVec3 fin
		local r3 = ReadVec3 fin
		local r4 = (ReadVec3 fin) * spnScale.value
		return (matrix3 r1 r2 r3 r4)
	)


	fn ReadColor fin =
	(
		local r = (ReadByte fin) / 255
		local g = (ReadByte fin) / 255
		local b = (ReadByte fin) / 255
		local a = (ReadByte fin) / 255
		return (color r g b a)
	)


	fn ReadStr fin length =
	(
		local str = ""
		for i = 1 to length do
		(
			local b = ReadByte fin
			if (str.count == 0 or b != 0) then
				str = str + (bit.intAsChar b)
		)
		str += (bit.intAsChar 0)
		return str
	)


	fn ReadFrameListData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    data\n    {\n"

		local frameCount = ReadLong fin
		for i = 1 to frameCount do
		(
			local matrix = ReadMatrix fin
			local parent = (ReadLong fin) + 1
			local unk0 = ReadLong fin

			append frameMatrix matrix
			append frameParent parent
		)

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)
	
	
	fn ReadFrame fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "      frame\n      {\n"

		local name = ReadStr fin sectionSize
		append frameName name

		if (chkDebug.checked) then
			format "      }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadFrameListExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    extension\n    {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadFrameList fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "  framelist\n  {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "  }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometryListData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    data\n    {\n"

		local geometryCount = ReadLong fin

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometryData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "      data\n      {\n"

		local flags = ReadShort fin
		local unk0 = ReadByte fin
		local disable = ReadByte fin
		local numFaces = ReadLong fin
		local numVerts = ReadLong fin
		local numMorphTargets = ReadLong fin
		local geomLength = 40 + ((bit.and flags rwOBJECT_VERTEX_COLOR) / 8) * numVerts * 4 + ((bit.and flags rwOBJECT_VERTEX_UV)  / 4) * numVerts * 8 + numFaces * 8 + numVerts * 12 + ((bit.and flags rwOBJECT_VERTEX_NORMAL) / 16) * numVerts * 12
		if (geomLength < sectionSize) then
		(
			local ambient = ReadFloat fin
			local diffuse = ReadFloat fin
			local specular = ReadFloat fin
		)
	
		local colorData = #()
		if (disable == 0) then
		(
			if ((bit.and flags rwOBJECT_VERTEX_COLOR) != 0) then
				for i = 1 to numVerts do
					append colorData (ReadColor fin)
		)
	
		local texcData = #()
		if (disable == 0) then
		(
			if ((bit.and flags rwOBJECT_VERTEX_UV) != 0) then
				for i = 1 to numVerts do
					append texcData (ReadVec2 fin)
		)
	
		local faceData = #()
		if (disable == 0) then
		(
			for i = 1 to numFaces do
			(
				local v2 = (ReadShort fin) + 1
				local v1 = (ReadShort fin) + 1
				local faceFlags = ReadShort fin
				local v3 = (ReadShort fin) + 1
				append faceData (point3 v1 v2 v3)
			)
		)

		local boundingSphereXPos = ReadFloat fin
		local boundingSphereYPos = ReadFloat fin
		local boundingSphereZPos = ReadFloat fin
		local boundingSphereRadius = ReadFloat fin
		local hasPosition = ReadLong fin
		local hasNormals = ReadLong fin

		local vertData = #()
		if (disable == 0) then
		(
			for i = 1 to numVerts do
				append vertData ((ReadVec3 fin) * spnScale.value)
		)
	
		local normData = #()
		if (disable == 0) then
		(
			if ((bit.and flags rwOBJECT_VERTEX_NORMAL) != 0) then
				for i = 1 to numVerts do
					append normData (ReadVec3 fin)
		)
	
		local myGeo = s_Geometry()
		myGeo.vertData = vertData
		myGeo.faceData = faceData
		myGeo.normData = normData
		myGeo.texcData = texcData
		myGeo.colorData = colorData
		append geomData myGeo

		if (chkDebug.checked) then
			format "      }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterialListData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "        data\n        {\n"

		local numMaterials = ReadLong fin

		sectionSize -= 4
		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "        }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterialData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "          data\n          {\n"

		local unk0 = ReadLong fin -- 0
		local unk1 = ReadLong fin -- -1
		local c = ReadColor fin
		local numTextures = ReadLong fin
		local unk2 = ReadFloat fin -- 1
		local unk3 = ReadFloat fin -- near 0
		local unk4 = ReadFloat fin -- -1

		if (chkDebug.checked) then
			format "          }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadTextureData fin sectionSize SectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "            data\n            {\n"

		local filterModeFlags = ReadShort fin
		local unk0 = ReadShort fin

		if (chkDebug.checked) then
			format "            }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadTextureString fin sectionSize SectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "            string\n            {\n"

		local name = ReadStr fin sectionSize
		if (chkDebug.checked) then
			print name

		if (chkDebug.checked) then
			format "            }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadTextureExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "            extension\n            {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "            }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadTexture fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "          texture\n          {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)		

		if (chkDebug.checked) then
			format "          }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterialExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "          extension\n          {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "          }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterial fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "        material\n        {\n"
		
		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)		

		if (chkDebug.checked) then
			format "        }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterialList fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "      materiallist\n      {\n"
		
		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)		

		if (chkDebug.checked) then
			format "      }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadMaterialSplit fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "        materialsplit\n        {\n"

		local unk0 = ReadLong fin
		local numSplits = (ReadLong fin) + 1
		local numFaces = ReadLong fin

		for i = 1 to numSplits do
		(
			local faceIndex = ReadLong fin
			local materialIndex = ReadLong fin
			for j = 1 to (faceIndex / 3) do
			(
				local v1 = ReadLong fin
				local v2 = ReadLong fin
				local v3 = ReadLong fin
			)
		)

		if (chkDebug.checked) then
			format "        }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometryExtensionUnknown fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "        <unknown 261>\n        {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "        }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometryExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "      extension\n      {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "      }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometry fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    geometry\n    {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadGeometryList fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "  geometrylist\n  {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "  }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadAtomicData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    data\n    {\n"
		local frameID = (ReadLong fin) + 1
		local geometryID = (ReadLong fin) + 1
		local unk0 = ReadLong fin
		local unk1 = ReadLong fin

		CreateAtomic geometryID frameID

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadAtomicExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "    extension\n    {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "    }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadAtomic fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "  atomic\n  {\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)		

		if (chkDebug.checked) then
			format "  }\n"

		return (fpointer - (ftell fin))
	)


	fn ReadClumpExtension fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "  extension\n  {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "  }\n"
		
		return (fpointer - (ftell fin))
	)
	

	fn ReadClumpData fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "  data\n  {\n"

		local sectEnd = (ftell fin) + sectionSize
		fseek fin sectEnd #seek_set

		if (chkDebug.checked) then
			format "  }\n"
		
		return (fpointer - (ftell fin))

	)


	fn ReadClump fin sectionSize sectionType =
	(
		local fpointer = (ftell fin)
		if (chkDebug.checked) then
			format "clump\n{\n"

		local endOfSection = (ftell fin) + sectionSize

		while ((ftell fin) < endOfSection) do
		(
			(ProcessChunk fin sectionType)
		)

		if (chkDebug.checked) then
			format "}\n"

		return (fpointer - (ftell fin))
	)


	fn ArrangeAtomics =
	(
		for i = 1 to frameMesh.count do
		(
			if (frameMesh[i] != undefined) then
			(
				if (frameParent[i] != 0) then
					frameMesh[i].parent = frameMesh[frameParent[i]]
				in coordsys parent frameMesh[i].rotation = conjugate (frameMatrix[i].rotationPart as quat)
				in coordsys parent frameMesh[i].position = frameMatrix[i].translationPart
			)
		)
	)

	fn ProcessChunk fin parentType =
	(
		local sectionType = ReadLong fin
		local sectionSize = ReadLong fin
		local versionInfo = ReadLong fin

		if (sectionSize > 0) then
		(
			case of
			(
				(sectionType == rwDATA):
				(
					case of
					(
						(parentType == rwFRAMELIST):	sectionSize -= (ReadFrameListData fin sectionSize sectionType)
						(parentType == rwGEOMETRYLIST):	sectionSize -= (ReadGeometryListData fin sectionSize sectionType)
						(parentType == rwCLUMP):		sectionSize -= (ReadClumpData fin sectionSize sectionType)
						(parentType == rwGEOMETRY):		sectionSize -= (ReadGeometryData fin sectionSize sectionType)
						(parentType == rwMATERIALLIST):	sectionSize -= (ReadMaterialListData fin sectionSize sectionType)
						(parentType == rwMATERIAL):		sectionSize -= (ReadMaterialData fin sectionSize sectionType)
						(parentType == rwTEXTURE):		sectionSize -= (ReadTextureData fin sectionSize sectionType)
						(parentType == rwATOMIC):		sectionSize -= (ReadAtomicData fin sectionSize sectionType)
						default:
						(
							if (chkDebug.checked) then
							(
								format "\n<Data default>\n"
								format "Location:\t\t%\n" (ftell fin)
								format "sectionType:\t%\n" sectionType
								format "sectionSize:\t%\n" sectionSize
								format "versionInfo:\t%\n\n" versionInfo
							)
							(fseek fin ((ftell fin) + sectionSize) #seek_set)
							sectionSize = 0
						)
					)
				)
				(sectionType == rwSTRING):				sectionSize -= (ReadTextureString fin sectionSize sectionType)
				(sectionType == rwEXTENSION):
				(
					case of
					(
						(parentType == rwCLUMP):		sectionSize -= (ReadClumpExtension fin sectionSize sectionType)
						(parentType == rwFRAMELIST):	sectionSize -= (ReadFrameListExtension fin sectionSize sectionType)
						(parentType == rwTEXTURE):		sectionSize -= (ReadTextureExtension fin sectionSize sectionType)
						(parentType == rwMATERIAL):		sectionSize -= (ReadMaterialExtension fin sectionSize sectionType)
						(parentType == rwGEOMETRY):		sectionSize -= (ReadGeometryExtension fin sectionSize sectionType)
						(parentType == rwATOMIC):		sectionSize -= (ReadAtomicExtension fin sectionSize sectionType)
						default:
						(
							if (chkDebug.checked) then
							(
								format "\n<Extension default>\n"
								format "Location:\t\t%\n" (ftell fin)
								format "sectionType:\t%\n" sectionType
								format "sectionSize:\t%\n" sectionSize
								format "versionInfo:\t%\n\n" versionInfo
							)
							(fseek fin ((ftell fin) + sectionSize) #seek_set)
							sectionSize = 0
						)
					)
				)
				(sectionType == rwTEXTURE):			sectionSize -= (ReadTexture fin sectionSize sectionType)
				(sectionType == rwMATERIAL):		sectionSize -= (ReadMaterial fin sectionSize sectionType)
				(sectionType == rwMATERIALLIST):		sectionSize -= (ReadMaterialList fin sectionSize sectionType)
				(sectionType == rwFRAMELIST):		sectionSize -= (ReadFrameList fin sectionSize sectionType)
				(sectionType == rwGEOMETRY):		sectionSize -= (ReadGeometry fin sectionSize sectionType)
				(sectionType == rwCLUMP):			sectionSize -= (ReadClump fin sectionSize sectionType)
				(sectionType == rwATOMIC):			sectionSize -= (ReadAtomic fin sectionSize sectionType)
				(sectionType == rwGEOMETRYLIST):		sectionSize -= (ReadGeometryList fin sectionSize sectionType)
				(sectionType == rwFRAME):			sectionSize -= (ReadFrame fin sectionSize sectionType)
				(sectionType == rwGEO_EXT_UNK):		sectionSize -= (ReadGeometryExtensionUnknown fin sectionSize sectionType)
				default:
				(
					if (chkDebug.checked) then
					(
						format "\n<Main default>\n"
						format "Location:\t\t%\n" (ftell fin)
						format "sectionType:\t%\n" sectionType
						format "sectionSize:\t%\n" sectionSize
						format "versionInfo:\t%\n\n" versionInfo
					)
					(fseek fin ((ftell fin) + sectionSize) #seek_set)
					sectionSize = 0
				)
			)
		)
		return true
	)


	fn ImportDFF fin =
	(
		frameName = #()
		frameMatrix = #()
		frameParent = #()
		geomData = #()
		frameMesh = #()

		(ProcessChunk fin 0)

		CreateDummies()
		ArrangeAtomics()

		return true
	)


	fn GetOpenFileStream =
	(
		local fname = getOpenFileName types:"DFF geometry (*.dff)|*.dff|All files (*.*)|*.*|"
		if (fname == undefined) then
			return undefined

		local fin = fopen fname "rb"
		if (fin == undefined) then
		(
			messageBox "Could not open file for binary reading!"
			return undefined
		)

		return fin
	)


	on btnImportNow pressed do
	(
		local oldCoordsys = set coordsys world
	
		local fin = GetOpenFileStream()		if (fin != undefined) then
		(
			if ((ImportDFF fin) == true) then
				print "DFF Import successful."
			else
				print "DFF Import failed."
			fclose fin
		)
	
		set coordsys oldCoordsys
	)
)

